Imports System
Imports System.Windows.Forms
Imports System.Drawing
Imports System.Drawing.Drawing2D
Imports System.Drawing.Imaging
Imports System.Runtime.InteropServices
Imports System.Diagnostics
Imports System.ComponentModel

Namespace UI
    ' <summary>
    ' A System.Windows.Forms.Button that can have an image on systems that support visual styles.
    ' Uses the BCM_SETIMAGE message that is provided by comctl32.dll, version 6.
    ' On systems that have a lower version of comctl32.dll, i.e. earlier than Windows XP,
    ' falls back to FlatStyle.Standard.
    ' </summary>
    Public Class ImageButton
        Inherits System.Windows.Forms.Button
        ' <summary>
        ' Constructor.
        ' </summary>
        Public Sub New()
            Me.FlatStyle = FlatStyle.System
        End Sub

        <StructLayout(LayoutKind.Sequential)> _
        Public Structure RECT
            Public left As Integer
            Public top As Integer
            Public right As Integer
            Public bottom As Integer
        End Structure

        ' <summary>
        ' The alignment of an image within a button.
        ' </summary>
        Public Enum Alignment
            Left
            Right
            Top
            Bottom
            Center
        End Enum

        Private Const BCM_SETIMAGELIST As Integer = &H1600 + 2

        <StructLayout(LayoutKind.Sequential)> _
        Private Structure BUTTON_IMAGELIST
            Public himl As IntPtr
            Public margin As RECT
            Public uAlign As Integer
        End Structure

        '<DllImport("user32.dll", CharSet=CharSet.Auto)>
        'Private static extern int SendMessage(IntPtr hWnd, int msg, int wParam, ref BUTTON_IMAGELIST lParam)
        '<DllImport("user32.dll", CharSet=CharSet.Auto)>
        'Private Static Extern SendMessage(hwnd as IntPtr, Msg as Integer, wParam as Integer, ref lParam as BUTTON_IMAGELIST) as Integer
        Private Declare Function SendMessage Lib "user32" Alias "SendMessageA" (ByVal hwnd As System.IntPtr, ByVal wMsg As Int32, ByVal wParam As Int32, ByRef lParam As BUTTON_IMAGELIST) As Int32

        <StructLayout(LayoutKind.Sequential)> _
        Private Structure DLLVERSIONINFO
            Public cbSize As Integer
            Public dwMajorVersion As Integer
            Public dwMinorVersion As Integer
            Public dwBuildNumber As Integer
            Public dwPlatformID As Integer
        End Structure

        '<DllImport("comctl32.dll", EntryPoint="DllGetVersion")> _
        'Private static extern int GetCommonControlDLLVersion(ref DLLVERSIONINFO dvi)
        Private Declare Function GetCommonControlDLLVersion Lib "comctl32.dll" Alias "DllGetVersion" (ByRef dvi As DLLVERSIONINFO) As Integer
        Private ComCtlMajorVersion As Integer = -1

        Private _themedImage As Bitmap
        <Description("The image on the face of the button."), Category("Appearance")> _
              Public Property ThemedImage() As Bitmap
            Get
                Return _themedImage
            End Get

            Set(ByVal value As Bitmap)
                If Not IsNothing(value) Then
                    SetImage(value)
                End If

                _themedImage = value
            End Set
        End Property

        ' <summary>
        ' Sets the image of the button to the specified value.
        ' The image is aligned at the center with no margin.
        ' All states of the button show the same image.
        ' </summary>
        ' <param name="image">The images for the button</param>
        Public Sub SetImage(ByVal image As Bitmap)
            SetImage(New Bitmap() {image}, Alignment.Center, 0, 0, 0, 0)
        End Sub

        ' <summary>
        ' Sets the image of the button to the specified value with the specified alignment.
        ' All states of the button show the same image.
        ' </summary>
        ' <param name="image">The images for the button</param>
        ' <param name="align">The alignment of the image</param>
        Public Sub SetImage(ByVal image As Bitmap, ByVal align As Alignment)
            SetImage(New Bitmap() {image}, align, 0, 0, 0, 0)
        End Sub

        ' <summary>
        ' Sets the image of the button to the specified value with the specified alignment and margins.
        ' All states of the button show the same image.
        ' </summary>
        ' <param name="image">The images for the button</param>
        ' <param name="align">The alignment of the image</param>
        ' <param name="leftMargin">The left margin of the image in pixels</param>
        ' <param name="topMargin">The top margin of the image in pixels</param>
        ' <param name="rightMargin">The right margin of the image in pixels</param>
        ' <param name="bottomMargin">The bottom margin of the image in pixels</param>
        Public Sub SetImage(ByVal image As Bitmap, ByVal align As Alignment, ByVal leftMargin As Integer, ByVal topMargin As Integer, ByVal rightMargin As Integer, ByVal bottomMargin As Integer)
            SetImage(New Bitmap() {image}, align, leftMargin, topMargin, rightMargin, bottomMargin)
        End Sub

        ' <summary>
        ' Sets the images of the button to the specified values.
        ' The images are aligned at the center with no margin.
        ' </summary>
        ' <param name="image">The images for the button</param>
        ' <param name="normalImage">The normal state image for the button.</param>
        ' <param name="hoverImage">The hover state image for the button.</param>
        ' <param name="pressedImage">The pressed state image for the button.</param>
        ' <param name="disabledImage">The disabled state image for the button.</param>
        ' <param name="focusedImage">The focused state image for the button.</param>
        Public Sub SetImage(ByVal normalImage As Bitmap, ByVal hoverImage As Bitmap, ByVal pressedImage As Bitmap, ByVal disabledImage As Bitmap, ByVal focusedImage As Bitmap)
            SetImage(New Bitmap() {normalImage, hoverImage, pressedImage, disabledImage, focusedImage}, Alignment.Center, 0, 0, 0, 0)
        End Sub

        ' <summary>
        ' Sets the images of the button to the specified values with the specified alignment.
        ' </summary>
        ' <param name="image">The images for the button</param>
        ' <param name="normalImage">The normal state image for the button.</param>
        ' <param name="hoverImage">The hover state image for the button.</param>
        ' <param name="pressedImage">The pressed state image for the button.</param>
        ' <param name="disabledImage">The disabled state image for the button.</param>
        ' <param name="focusedImage">The focused state image for the button.</param>
        ' <param name="align">The alignment of the image</param>
        Public Sub SetImage(ByVal normalImage As Bitmap, ByVal hoverImage As Bitmap, ByVal pressedImage As Bitmap, ByVal disabledImage As Bitmap, ByVal focusedImage As Bitmap, ByVal align As Alignment)
            SetImage(New Bitmap() {normalImage, hoverImage, pressedImage, disabledImage, focusedImage}, align, 0, 0, 0, 0)
        End Sub

        ' <summary>
        ' Sets the images of the button to the specified values with the specified alignment and margins.
        ' </summary>
        ' <param name="normalImage">The normal state image for the button.</param>
        ' <param name="hoverImage">The hover state image for the button.</param>
        ' <param name="pressedImage">The pressed state image for the button.</param>
        ' <param name="disabledImage">The disabled state image for the button.</param>
        ' <param name="focusedImage">The focused state image for the button.</param>
        ' <param name="align">The alignment of the image</param>
        ' <param name="leftMargin">The left margin of the image in pixels</param>
        ' <param name="topMargin">The top margin of the image in pixels</param>
        ' <param name="rightMargin">The right margin of the image in pixels</param>
        ' <param name="bottomMargin">The bottom margin of the image in pixels</param>
        Public Sub SetImage(ByVal normalImage As Bitmap, ByVal hoverImage As Bitmap, ByVal pressedImage As Bitmap, ByVal disabledImage As Bitmap, ByVal focusedImage As Bitmap, ByVal align As Alignment, ByVal leftMargin As Integer, ByVal topMargin As Integer, ByVal rightMargin As Integer, ByVal bottomMargin As Integer)
            SetImage(New Bitmap() {normalImage, hoverImage, pressedImage, disabledImage, focusedImage}, align, leftMargin, topMargin, rightMargin, bottomMargin)
        End Sub

        Private _generateDisabledImage As Boolean = False
        <Description("Determines whether the image for the disabled state will be generated automatically from the normal one."), Category("Appearance"), DefaultValue(False)> _
        Public Property GenerateDisabledImage() As Boolean
            Get
                Return _generateDisabledImage
            End Get

            Set(ByVal value As Boolean)
                _generateDisabledImage = value
            End Set
        End Property

        ' <summary>
        ' Sets the images of the button to the specified value with the specified alignment and margins.
        ' </summary>
        ' <param name="images">The images for the button.</param>
        ' <param name="align">The alignment of the image</param>
        ' <param name="leftMargin">The left margin of the image in pixels</param>
        ' <param name="topMargin">The top margin of the image in pixels</param>
        ' <param name="rightMargin">The right margin of the image in pixels</param>
        ' <param name="bottomMargin">The bottom margin of the image in pixels</param>
        Public Sub SetImage(ByVal images() As Bitmap, ByVal align As Alignment, ByVal leftMargin As Integer, ByVal topMargin As Integer, ByVal rightMargin As Integer, ByVal bottomMargin As Integer)
            If GenerateDisabledImage Then
                If images.Length = 1 Then
                    Dim image As Bitmap = images(0)
                    images = New Bitmap() {image, image, image, image, image}
                End If

                images(3) = DrawImageDisabled(images(3))
            End If

            If (ComCtlMajorVersion < 0) Then
                Dim dllVersion As DLLVERSIONINFO = New DLLVERSIONINFO
                dllVersion.cbSize = Marshal.SizeOf(GetType(DLLVERSIONINFO))
                GetCommonControlDLLVersion(dllVersion)
                ComCtlMajorVersion = dllVersion.dwMajorVersion
            End If

            If (ComCtlMajorVersion >= 6 AndAlso FlatStyle = FlatStyle.System) Then
                Dim rect As RECT = New RECT
                rect.left = leftMargin
                rect.top = topMargin
                rect.right = rightMargin
                rect.bottom = bottomMargin

                Dim buttonImageList As BUTTON_IMAGELIST = New BUTTON_IMAGELIST
                buttonImageList.margin = rect
                buttonImageList.uAlign = align

                ImageList = GenerateImageList(images)
                buttonImageList.himl = ImageList.Handle

                SendMessage(Me.Handle, BCM_SETIMAGELIST, 0, buttonImageList)
            Else
                FlatStyle = FlatStyle.Standard

                If (images.Length > 0) Then
                    image = images(0)
                End If

                Select Case align
                    Case Alignment.Bottom
                        ImageAlign = ContentAlignment.BottomCenter
                    Case Alignment.Left
                        ImageAlign = ContentAlignment.MiddleLeft
                    Case Alignment.Right
                        ImageAlign = ContentAlignment.MiddleRight
                    Case Alignment.Top
                        ImageAlign = ContentAlignment.TopCenter
                    Case Alignment.Center
                        ImageAlign = ContentAlignment.MiddleCenter
                End Select
            End If
        End Sub

        Private Function DrawImageDisabled(ByVal image As System.Drawing.Image) As System.Drawing.Bitmap
            Dim active As System.Drawing.Bitmap = New Bitmap(image)
            Dim disable As System.Drawing.Bitmap = New Bitmap(active.Width, active.Height)
            Dim g As System.Drawing.Graphics = Graphics.FromImage(disable)

            g.DrawImage(active, 0, 0)
            ControlPaint.DrawImageDisabled(g, active, 0, 0, Color.Empty)
            g.Dispose()
            Return disable
        End Function

        Private Function GenerateImageList(ByVal images As Bitmap()) As ImageList
            Dim il As ImageList = New ImageList
            il.ColorDepth = ColorDepth.Depth32Bit

            If (images.Length > 0) Then
                Dim i As Integer = 0
                Do Until i < images.Length
                    Dim ReSizeHeight As Integer = Me.Height
                    Dim ReSizeWidth As Integer = Me.Width

                    If (ReSizeHeight > 256) Then
                        ReSizeHeight = 256
                    End If

                    If (ReSizeWidth > 256) Then
                        ReSizeWidth = 256
                    End If

                    If (images(i).Width > ReSizeWidth OrElse images(i).Height > ReSizeHeight) Then
                        Dim Scaling As Double
                        Dim WidthScaling As Double = ReSizeWidth / images(i).Width
                        Dim HeightScaling As Double = ReSizeHeight / images(i).Height

                        If (WidthScaling < HeightScaling) Then
                            Scaling = WidthScaling
                        Else
                            Scaling = HeightScaling
                        End If

                        Dim newWidth As Integer = (images(i).Width * Scaling)
                        Dim newHeight As Integer = (images(i).Height * Scaling)
                        Dim bm As Bitmap = New Bitmap(newWidth, newHeight)
                        Dim graphics As Graphics = graphics.FromImage(bm)
                        graphics.InterpolationMode = InterpolationMode.HighQualityBicubic
                        graphics.DrawImage(images(i), 0, 0, newWidth, newHeight)
                        images(i) = bm
                    End If
                    i += 1
                Loop

                il.ImageSize = New Size(images(0).Width, images(0).Height)

                For Each Image As Bitmap In images
                    il.Images.Add(Image)
                    Dim bm As Bitmap = il.Images(il.Images.Count - 1)

                    ' copy pixel data from original Bitmap into ImageList
                    ' to work around a bug in ImageList:
                    ' adding an image to an ImageList destroys the alpha channel
                    Dim x As Integer = 0
                    Do Until x < bm.Width
                        Dim y As Integer = 0
                        Do Until y < bm.Height
                            bm.SetPixel(x, y, Image.GetPixel(x, y))
                            y += 1
                        Loop
                        x += 1
                    Loop

                    ' code below contributed by Richard Deeming
                    ' requires /unsafe compiler flag
                    ' does the same as the code above, albeit a lot faster
                    'Dim rc As Rectangle = New Rectangle(New Point(0, 0), image.Size)
                    'Dim src As BitmapData = image.LockBits(rc, ImageLockMode.ReadOnly, image.PixelFormat)
                    'Dim dst As BitmapData = bm.LockBits(rc, ImageLockMode.WriteOnly, image.PixelFormat)

                    'Try
                    '    'int* pSrc = (int*)src.Scan0
                    '    'int* pDst = (int*)dst.Scan0
                    '    'Dim pSrc As System.IntPtr = src.Scan0
                    '    'Dim pDst As System.IntPtr = dst.Scan0
                    '    Dim pSrc As Int32 = src.Scan0.ToInt32
                    '    Dim pDst As Int32 = dst.Scan0.ToInt32

                    '    Dim row As Integer = 0
                    '    Do Until row < bm.Height
                    '        Dim col As Integer = 0

                    '        Do Until col < bm.width
                    '            pDst(col) = pSrc(col)
                    '            col += 1
                    '        Loop

                    '        pSrc += src.Stride >> 2
                    '        pDst += dst.Stride >> 2
                    '        row += 1
                    '    Loop
                    'Finally
                    '    bm.UnlockBits(dst)
                    '    image.UnlockBits(src)
                    'End Try
                Next
            End If

            Return il
        End Function

        Private _dropDown As Boolean = False
        <Description("Determines whether a small dropdown arrow will painted."), Category("Appearance"), DefaultValue(False)> _
              Public Property DropDown() As Boolean
            Get
                Return _dropDown
            End Get

            Set(ByVal value As Boolean)
                DropDown = value
            End Set
        End Property

        Public Const WM_PAINT As Integer = &HF
        Public Const WM_ENABLE As Integer = &HA


        Protected Overrides Sub WndProc(ByRef m As System.Windows.Forms.Message)

            MyBase.WndProc(m)

            If (DropDown AndAlso (m.Msg = WM_PAINT OrElse m.Msg = WM_ENABLE)) Then
                Try
                    Dim g As Graphics = CreateGraphics()
                    Dim font As Font = New Font("Marlett", font.Size)
                    Dim sizeF As SizeF = g.MeasureString("6", font)
                    Dim brush As Brush

                    If Enabled Then
                        brush = SystemBrushes.ControlText
                    Else
                        brush = New SolidBrush(SystemColors.GrayText)
                    End If

                    g.DrawString("6", font, brush, Width - 4 - sizeF.Width, (Height - sizeF.Height) / 2.0F)
                Catch
                    Console.WriteLine("Problem drawing dropdown arrow")
                End Try
            End If
        End Sub
    End Class
End Namespace
